import sys

from flask import Flask

from .config import GH_TOKEN, HOST, PORT, HIDDEN
from . import routes


def main():
    app = routes.setup(Flask(HIDDEN[1:-1]))
    app.run(host=HOST, port=PORT)


if __name__ == '__main__':
    if GH_TOKEN is None:
        sys.exit('GH_TOKEN must be set.')
    main()



# -*- coding: utf-8 -*-
from os import environ as env

# Identifier used to recognize auto-generated comments
HIDDEN = '<KGB>'
STATUS_CONTEXT = 'Комитет Git\'a Безопасности'

MAX_STATUS_LENGTH = int(env.get('MAX_STATUS_LENGTH') or 50)
MAX_LINE_LENGTH = int(env.get('MAX_LINE_LENGTH') or 72)

_DEFAULT_TYPES = ','.join([
    'chore',
    'doc',
    'feat',
    'fix',
    'perf',
    'refactor',
    'revert',
    'style',
    'test',
    'version'])
TYPES = frozenset((env.get('AUTHORIZED_TYPES') or _DEFAULT_TYPES).split(','))

GH_TOKEN = env.get('GH_TOKEN')

HOST = env.get('KGB_HOST') or '0.0.0.0'
PORT = env.get('KGB_PORT') or 5000

INFO = 'https://github.com/sagacify/komitet-gita-bezopasnosti'



"""Github API module."""
import logging

import requests

from .config import GH_TOKEN
from .config import HIDDEN
from .config import INFO
from .config import STATUS_CONTEXT

log = logging.getLogger(__name__)


def protect(string):
    return string.replace('[', r'\[')


def quote(string):
    return '>' + string.replace('\n', '\n>')


def get_commits(url):
    commits = requests.get(url, params={'access_token': GH_TOKEN})
    return commits.json()


def get_comments(url):
    comments = requests.get(url, params={'access_token': GH_TOKEN})
    result = []
    for comment in comments.json():
        if comment['body'].startswith(HIDDEN):
            result.append(comment)
    return result


def delete_comment(comment):
    requests.delete(comment['url'], params={'access_token': GH_TOKEN})


def upsert_comment(url, messages):
    comments = get_comments(url)
    message = HIDDEN + '\n\n'.join(messages) + 'More info [here](%s)' % INFO

    if len(messages) == 0:
        for comment in comments:
            delete_comment(comment)
    else:
        if len(comments) > 1:
            for comment in comments:
                delete_comment(comment)
            create_comment(url, message)
        if len(comments) == 1:
            if comments[0]['body'] != message:
                update_comment(comments[0], message)
        else:
            create_comment(url, message)


def update_comment(comments, body):
    requests.patch(comments['url'],
                   params={'access_token': GH_TOKEN},
                   json={'body': body})


def create_comment(url, body):
    requests.post(url,
                  params={'access_token': GH_TOKEN},
                  json={'body': body})


def update_status(url, state, message):
    requests.post(url,
                  params={'access_token': GH_TOKEN},
                  json={'state': state,
                        'context': STATUS_CONTEXT,
                        'description': message})



r"""Rule module.

This module and sub-modules contain all the rules that can be applied
to the commit messages.


>>> good_message = '''fix(kgb): Make kgb reliable
...
... We used docstrings to keep the tests as close as possible to the codes.
... Look! That line was exactly 72 characters long!!!
... Yes I know that came at the price of an extra s at code.
... # This\tline should be ignored as it is a comment. So any violation should
... '''

>>> bad_message = '''fixes(k g b): Make it quick and dirty, make it bads
... We try to make explanations as understandable as possible but we have some
... \tissues with code consistency.'''

>>> len(apply_rules(good_message))
0
>>> len(apply_rules(bad_message))
5
"""

import re
from inspect import getmembers, isfunction
from . import line_rules
from . import status_rules


def _get_rules(module):
    """Get all functions from a module."""
    return [tup[1] for tup in getmembers(module, isfunction)
            if tup[0][0] != '_']


_line_rules = _get_rules(line_rules)
_status_rules = _get_rules(status_rules)


def _is_merge(status_line):
    return status_line.startswith('Merge pull request #')


def split_lines(commit_message):
    return re.split('\r?\n', commit_message)


def apply_rules(commit_message):
    """Apply all rules to the commit message."""
    errors = []
    if _is_merge(commit_message):
        return errors
    commit_lines = split_lines(commit_message)
    status = commit_lines[0]
    commit_lines = [l for l in commit_lines if not l.startswith('#')]
    for rule in _status_rules:
        err = rule(status)
        if err is not None:
            errors.append(err)
    for check in _line_rules:
        err = check(commit_lines)
        if err is not None:
            errors.append(err)
    return errors



import re

from ..config import MAX_STATUS_LENGTH
from ..config import TYPES

STATUS_PATTERN = re.compile(
    r'^([^()\s]*)(\s*)\(([^()\s]*)\)(\s*):(\s*)(.*\S)(\s*)$')

PRETTY_TYPES = ', '.join(sorted(TYPES))


def check_status_length(status_line):
    """Check status length.

    >>> check_status_length('this is short enough')

    >>> check_status_length('TOO_LONG' * 10)
    'Shorten status from {0} to {1} characters'
    """.format(len('TOO_LONG' * 10), MAX_STATUS_LENGTH)
    if len(status_line) > MAX_STATUS_LENGTH:
        return 'Shorten status from {0} to {1} characters'.format(
            len(status_line), MAX_STATUS_LENGTH)


def check_status_formatting(status_line):
    r"""Check status formatting.
    >>> check_status_formatting('fix(this): This is good')

    >>> print(check_status_formatting('fixing(this): This is not good'))
    Begin status with one of the following types:
    *chore, doc, feat, fix, perf, refactor, revert, style, test, version*

    >>> print(check_status_formatting('testing(this): This is not good'))
    Begin status with one of the following types:
    *chore, doc, feat, fix, perf, refactor, revert, style, test, version*

    >>> print(check_status_formatting('fix(th)is): This is not good'))
    Conform status to `<type>(<scope>): <subject>` pattern
    type and scope cannot contain spaces or parenthesis

    >>> print(check_status_formatting('fix(th is): This is not good'))
    Conform status to `<type>(<scope>): <subject>` pattern
    type and scope cannot contain spaces or parenthesis

    >>> print(check_status_formatting('fix (this) :   This is not good'))
    Remove space between type and scope.
    Remove space between scope and colon.
    Use a single space between colon and subject.

    >>> print(check_status_formatting('fix(this):n  '))
    Add space after colon.
    Make subject at least three character long.
    Strip trailing whitespaces from status line.

    >>> check_status_formatting('fix(this): Allow final non-word character.')

    """
    match = STATUS_PATTERN.match(status_line)
    if match is None:
        return ('Conform status to `<type>(<scope>): <subject>` pattern\n' +
                'type and scope cannot contain spaces or parenthesis')
    errors = []
    if match.group(1) not in TYPES:
        errors.append('Begin status with one of the following types:\n' +
                      '*%s*' % PRETTY_TYPES)
    if len(match.group(2)) > 0:
        errors.append('Remove space between type and scope.')
    if len(match.group(4)) > 0:
        errors.append('Remove space between scope and colon.')
    if match.group(5) == '':
        errors.append('Add space after colon.')
    elif match.group(5) != ' ':
        errors.append('Use a single space between colon and subject.')
    if len(match.group(6).strip()) < 3:
        errors.append('Make subject at least three character long.')
    if len(match.group(7)) > 0:
        errors.append('Strip trailing whitespaces from status line.')
    if len(errors) == 0:
        return None
    else:
        return '\n'.join(errors)



import logging
import flask
from . import github
from . import rules

log = logging.getLogger(__name__)


def check_pr(commits_url, comments_url, status_url):
    github.update_status(status_url, 'pending',
                         'The KGB is reviewing your commits.')
    all_errors = 0
    messages = []
    for commit in github.get_commits(commits_url):
        message = commit['commit']['message']
        errors = rules.apply_rules(message)
        if len(errors) > 0:
            sha = commit['sha']
            messages.append('\n'.join([sha, github.quote(message), ''] +
                                      errors))
            all_errors += len(errors)

    log.debug(all_errors)
    github.upsert_comment(comments_url, messages)
    if all_errors == 0:
        github.update_status(status_url, 'success', 'We are proud of you!')
    elif all_errors == 1:
        github.update_status(status_url, 'error', 'Fix the error!')
    else:
        github.update_status(status_url, 'error',
                             'Fix the {} errors!'.format(all_errors))


def github_pr():
    body = flask.request.json
    if body is None:
        return '', 500
    keys = body.keys()
    if 'zen' in keys:
        # This is the initial event sent to test the hook.
        return '', 204
    if 'action' not in keys or 'pull_request' not in keys:
        log.error({'error': 'not_a_pull_request', 'body': body})
        return '', 500
    elif body['action'] in {'assigned', 'unassigned',
                            'labeled', 'unlabeled', 'closed'}:
        log.debug({'event': 'non_event', 'body': body})
        return '', 200
    elif body['action'] in {'opened', 'reopened', 'synchronize'}:
        log.debug({'event': 'analyse_pr', 'body': body})
        check_pr(body['pull_request']['commits_url'],
                 body['pull_request']['comments_url'],
                 body['pull_request']['statuses_url'])
        return ''

    return '', 204


def setup(app):
    app.add_url_rule('/github/pr',
                     'github_pr',
                     github_pr,
                     methods=['POST'])
    return app



from ..config import MAX_LINE_LENGTH


def check_second_line(commit_lines):
    r"""Check second line is blank line and there is a body.

    >>> from . import split_lines
    >>> check_second_line(split_lines('first line\n\nbody'))

    >>> check_second_line(split_lines('first line\nbody'))
    'Separate status and body with an empty line.'
    """
    if len(commit_lines) > 1 and len(commit_lines[1]) > 0:
        return 'Separate status and body with an empty line.'


def check_body_line_length(commit_lines):
    """Check line length.

    >>> check_body_line_length(["", "", 'this is short enough'])

    >>> check_body_line_length(["", "", 'TOO_LONG' * 10])
    'Wrap body lines at 72 characters'
    """
    for line in commit_lines:
        if len(line) > MAX_LINE_LENGTH:
            return 'Wrap body lines at {0} characters'.format(MAX_LINE_LENGTH)


def check_for_tabs(commit_lines):
    r"""Check for tabs.

    >>> check_for_tabs(['This is a good commit.'])

    >>> check_for_tabs(['This is a\tbad commit.'])
    'Use soft tabs!'
    """
    for line in commit_lines:
        if line.find('\t') != -1:
            return 'Use soft tabs!'



from distutils.core import setup
import io
import os


readme = 'README.md'
if os.path.isfile('./README.rst'):
    readme = 'README.rst'

print('Using %s as readme' % readme)
with io.open(readme, encoding='utf-8') as file:
    readme = file.read()

setup(
    name='kgitb',
    packages=['kgitb', 'kgitb.rules'],
    scripts=['bin/resident'],
    version='0.1.4',
    description='A commit message linter',
    long_description=readme,
    author='Augustin Borsu',
    author_email='dev@sagacify.com',
    url='https://github.com/Sagacify/komitet-gita-bezopasnosti',
    extras_require={
        'WEB': ['flask', 'requests']
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: ' +
        'GNU General Public License v3 or later (GPLv3+)',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Topic :: Text Processing'
    ],
    keywords='git lint russian_history'
)
